angular.module('starter.services')

.factory('IDBuilder', function(IDs, SavedData, ImageEffects, $q) {
  return {
    build: function(jsonFileName) {

      var defer = $q.defer();

      IDs.get(jsonFileName).then(jsonFileReturned, error);

      function jsonFileReturned(data) {
        var canvas, ctx;

        //get json data for selected template, and inputted user info
        var IDdata = data;
        var savedData = SavedData.getInputs();

        //load bg image first and wait for it to load
        var bg = new Image();
        bg.onload = function() {
          //make a canvas the same size as bg and draw bg into it
          canvas = document.createElement('canvas');
          canvas.width = bg.width;
          canvas.height = bg.height;
          ctx = canvas.getContext('2d');
          ctx.drawImage(bg, 0, 0);

          //load images and extras, apply effects, then draw everything in order and save
          loadImages()
          .then(applyEffects)
          .then(function() {
            drawImages();
            drawRects();
            drawTexts();
            drawExtras();
            save();
          });

        };
        bg.src = "img/templates/bgs/" + IDdata.bgFilename;

        //loads all images
        function loadImages() {
          var defer = $q.defer();
          var images = IDdata.rendering.images || [];
          var extras = IDdata.rendering.extras || [];
          var promises = [];
          for (var i in images) {
            promises.push(loadImage(images[i]));
          }
          for (var j in extras) {
            promises.push(loadExtra(extras[j]));
          }
          $q.all(promises).then(function() {
            defer.resolve();
          });
          return defer.promise;
        }

        function loadImage(render) {
          var defer = $q.defer();
          var img = new Image();
          img.onload = function() {
            render.img = img;
            defer.resolve();
          };
          if (render.type == 'photo') {
            img.src = savedData.photo || 'img/blue_outline.png';
          } else if (render.type == 'signature') {
            img.src = savedData.signaturePNG || 'img/nothing.png';
          }
          return defer.promise;
        }

        function loadExtra(render) {
          var defer = $q.defer();
          var img = new Image();
          img.onload = function() {
            render.img = img;
            defer.resolve();
          };
          img.src = render.path;
          return defer.promise;
        }

        function applyEffects() {
          var defer = $q.defer();
          var renders = IDdata.rendering.images;
          var promises = [];
          for (var i in renders) {
            if (renders[i].effect == 'grayscale') {
              promises.push(ImageEffects.grayscale(renders[i]));
            } else if (renders[i].effect == 'colorFilter') {
              promises.push(ImageEffects.colorFilter(renders[i], IDdata.paints[renders[i].paint].color));
            } else if (renders[i].effect == 'invertColor') {
              promises.push(ImageEffects.invertColor(renders[i]));
            }
          }
          $q.all(promises).then(function() {
            defer.resolve();
          });
          return defer.promise;
        }

        function drawImages() {
          for (var i in IDdata.rendering.images) {
            var render = IDdata.rendering.images[i];
            var paint = IDdata.paints[render.paint] || {};
             drawImage(render, paint);
          }
        }

        function drawExtras() {
          for (var i in IDdata.rendering.extras) {
            var extra = IDdata.rendering.extras[i];
            var paint = IDdata.paints[extra.paint] || {};
            var img = extra.img;
            drawImage(extra, paint, img);
          }
        }

        function drawTexts() {
          for (var i in IDdata.rendering.text) {
            var text = IDdata.rendering.text[i];
            var paint = IDdata.paints[text.paint] || {};
            drawText(text, paint, text.name);
          }
        }

        function drawRects() {
          for (var i in IDdata.rendering.rects) {
            var rect = IDdata.rendering.rects[i];
            var paint = IDdata.paints[rect.paint] || {};
            drawRect(rect, paint);
          }
        }

        function drawImage(render, paint) {
          ctx.save();
          ctx.globalAlpha = paint.alpha || 1;
          ctx.translate(render.x, render.y);
          ctx.rotate((render.angle || 0) * Math.PI/180);
          ctx.translate(-render.x, -render.y);
          if (render.w && render.h) {
            ctx.drawImage(render.img, render.x, render.y, render.w, render.h);
          } else {
            ctx.drawImage(render.img, render.x, render.y);
          }
          ctx.restore();
        }

        function drawText(render, paint, index) {
          var text;
          if (render.type === 'disclaimer') {
            text = "Fake ID Generator";
          } else {
            text = savedData.texts[index] || "";
            var label = render.label || "";
            var after = render.after || "";
            text = label + text + after;
            if (render.related) {
              for (var name in render.related) {
                var related = render.related[name];
                text += (related.label || "") + (savedData.texts[related.name] || "") + (related.after || "");
              }
            }
          }
          ctx.save();
          ctx.font = paint.font;
          ctx.fillStyle = paint.color;
          ctx.lineWidth = paint.strokeWidth;
          ctx.strokeStyle = paint.strokeColor;
          ctx.shadowColor = paint.shadowColor || 'black';
          ctx.shadowBlur = paint.shadowBlur || 0;
          ctx.shadowOffsetX = paint.shadowOffsetX || 0;
          ctx.shadowOffsetY = paint.shadowOffsetY || 0;
          ctx.textAlign = paint.align;
          ctx.globalAlpha = paint.alpha || 1;
          var scaleBack = render.x / (render.x * (paint.scaleX || 1)); //to correct x,y after scaling w,h - will be 1 if no scaleX given
          ctx.scale((paint.scaleX || 1), 1);
          if (paint.color) ctx.fillText(text, render.x * scaleBack, render.y);
          if (paint.strokeColor) ctx.strokeText(text, render.x * scaleBack, render.y);
          if (paint.underlineColor) {
            ctx.fillStyle = paint.underlineColor;
            ctx.fillRect(render.x, render.y + 5, ctx.measureText(text).width, 2);
          }

          ctx.restore();
        }

        function drawRect(render, paint) {
          ctx.save();
          ctx.rect(render.x, render.y, render.w, render.h);
          ctx.fillStyle = paint.fillColor || "black";
          ctx.strokeStyle = paint.strokeColor || "black";
          ctx.lineWidth = paint.strokeWidth || 1;
          ctx.globalAlpha = paint.alpha || 1;
          if (paint.fillColor) ctx.fill();
          if (paint.strokeColor) ctx.stroke();
          ctx.restore();
        }

        function save() {
          defer.resolve(canvas.toDataURL("image/jpeg", 1));
        }

      }

      function error(reason) {
        console.error(reason);
      }

      return defer.promise;
    }
  };
});
